import logging
import traceback
import os

from amplitude import Amplitude, BaseEvent
from typing import Optional

import requests

from tonic_textual.classes.httpclient import HttpClient

AMPLITUDE_API_KEY = os.getenv(
    "TONIC_TEXTUAL_ANALYTICS_KEY", "d04b22d8940cf7a21c5c86a803f3c709"
)


class TextualTelemetry:
    def __init__(
        self, base_url: str, api_key: Optional[str] = None, verify: bool = True
    ):
        try:
            if api_key is None:
                api_key = os.environ.get("TONIC_TEXTUAL_API_KEY")
                if api_key is None:
                    raise Exception(
                        "No API key provided. Either provide an API key, or set the API "
                        "key as the value of the TEXTUAL_API_KEY environment "
                        "variable."
                    )
            self.api_key = api_key
            self.client = HttpClient(base_url, self.api_key, verify)
            self.verify = verify

            with requests.Session() as session:
                response = self.client.http_get("/api/accounts", session=session)
                self.user_id = response.get("userName", "sdk_unknown_user")

            self.telemetry_client = Amplitude(AMPLITUDE_API_KEY)
            self.telemetry_client.configuration.flush_max_retries = 5
            self.telemetry_client.configuration.logger = logging.getLogger(__name__)
            self.telemetry_client.configuration.server_zone = "US"

            self.enabled = os.getenv("TONIC_TEXTUAL_SDK_TELEMETRY", "True").lower() in (
                "true",
                "1",
                "t",
            )
            self.telemetry_client.configuration.opt_out = not self.enabled
        except Exception as _:
            self.enabled = False

    def log_function_call(self):
        try:
            if not self.enabled:
                return

            # get the function call from the stack
            stack = traceback.extract_stack()

            called_function, parent_caller = stack[-2], stack[-3]

            if "tonic_textual" in parent_caller.filename:
                return
            else:
                # send amplitude event if function was directly called by user code
                event = BaseEvent(
                    event_type="SDK Call",
                    event_properties={"function_name": called_function.name},
                    user_id=self.user_id,
                )
                self.telemetry_client.track(event)
        except Exception as _:
            return



from typing import Optional


class DatasetFile:
    """
    Class to store the metadata for a dataset file.

    Parameters
    ----------
    id : str
        The identifier of the dataset file.

    name: str
        The file name of the dataset file.

    num_rows : long
        The number of rows in the dataset file.

    num_columns: int
        The number of columns in the dataset file.

    processing_status: string
        The status of the dataset file in the processing pipeline. Possible values are
        'Completed', 'Failed', 'Cancelled', 'Running', and 'Queued'.

    processing_error: string
        If the dataset file processing failed, a description of the issue that caused
        the failure.

    uploaded_timestamp: str
        Timestamp in UTC when dataset file was uploaded to the dataset.
    """

    def __init__(
        self,
        id: str,
        name: str,
        num_rows: Optional[int],
        num_columns: int,
        processing_status: str,
        processing_error: Optional[str],
    ):
        self.id = id
        self.name = name
        self.num_rows = num_rows
        self.num_columns = num_columns
        self.processing_status = processing_status
        self.processing_error = processing_error

    def describe(self) -> str:
        """Returns the dataset file metadata as string. Includes the identifier, file
        name, number of rows, and number of columns."""
        description = f"File: {self.name} [{self.id}]\n"
        description += f"Number of rows: {self.num_rows}\n"
        description += f"Number of columns: {self.num_columns}\n"
        description += f"Status: {self.processing_status}\n"
        if self.processing_status != "" and self.processing_error is not None:
            description += f"Error: {self.processing_error}\n"
        return description



import os
from typing import List, Optional, Union

import requests
from requests import RequestException

from tonic_textual.classes.httpclient import HttpClient
from tonic_textual.classes.pipeline import Pipeline
from tonic_textual.classes.tonic_exception import PipelineDeleteError, PipelineCreateError

from tonic_textual.classes.textual_telemetry import TextualTelemetry


class TonicTextualParse:
    """Wrapper class for invoking Tonic Textual API

    Parameters
    ----------
    base_url : str
        The URL to your Tonic Textual instance. Do not include trailing backslashes.
    api_key : str
        Your API token. This argument is optional. Instead of providing the API token
        here, it is recommended that you set the API key in your environment as the
        value of TEXTUAL_API_KEY.
    verify: bool
        Whether SSL Certification verification is performed.  This is enabled by
        default.
    Examples
    --------
    >>> from tonic_textual.parse_api import TonicTextualParse
    >>> textual = TonicTextualParse("https://textual.tonic.ai")
    """

    def __init__(
        self, base_url: str, api_key: Optional[str] = None, verify: bool = True
    ):
        if api_key is None:
            api_key = os.environ.get("TONIC_TEXTUAL_API_KEY")
            if api_key is None:
                raise Exception(
                    "No API key provided. Either provide an API key, or set the API "
                    "key as the value of the TEXTUAL_API_KEY environment "
                    "variable."
                )
        self.api_key = api_key
        self.client = HttpClient(base_url, self.api_key, verify)
        self.verify = verify
        self.telemetry_client = TextualTelemetry(base_url, api_key, verify)

    def get_pipelines(self) -> List[Pipeline]:
        """Get the pipelines for the Tonic Textual instance.

        Returns
        -------
        List[Pipeline]
            A list of pipeline objects, ordered by their creation timestamp.
        Examples
        --------
        >>> latest_pipeline = textual.get_pipelines()[-1]
        """
        self.telemetry_client.log_function_call()

        with requests.Session() as session:
            response = self.client.http_get("/api/parsejobconfig", session=session)
            pipelines: List[Pipeline] = []
            for x in response:
                pipelines.append(Pipeline(x["name"], x["id"], self.client))
            return pipelines

    def create_pipeline(self, pipeline_name: str) -> Pipeline:
        """Create a new pipeline.

        Parameters
        ----------
        pipeline_name: str
            The name of the pipeline.

        Returns
        -------
        Pipeline
            The newly created pipeline.
        """
        self.telemetry_client.log_function_call()

        try:
            p = self.client.http_post(f"/api/parsejobconfig/local-files", data={"name": pipeline_name})
            return Pipeline(p.get('name'), p.get('id'), self.client)
        except RequestException as req_err:
            if hasattr(req_err, 'response') and req_err.response is not None:
                status_code = req_err.response.status_code
                error_message = req_err.response.text
                raise PipelineCreateError(f"Error {status_code}: {error_message}")
            else:
                raise req_err


    def delete_pipeline(self, pipeline_id: str):
        """Delete a pipeline.


        Parameters
        ----------
        pipeline_id: str
            The ID of the pipeline.
        """
        self.telemetry_client.log_function_call()

        try:
            result = self.client.http_delete(f"/api/parsejobconfig/{pipeline_id}")
            return result
        except RequestException as req_err:
            if hasattr(req_err, 'response') and req_err.response is not None:
                status_code = req_err.response.status_code
                error_message = req_err.response.text
                raise PipelineDeleteError(f"Error {status_code}: {error_message}")
            else:
                raise req_err

    def get_pipeline_by_id(self, pipeline_id: str) -> Union[Pipeline, None]:
        """Get the pipeline by ID.

        Parameters
        ----------
        pipeline_id: str
            The ID of the pipeline.

        Returns
        -------
        Union[Pipeline, None]
            The Pipeline object or None if no pipeline is found.
        """
        self.telemetry_client.log_function_call()

        pipelines = self.get_pipelines()
        found_pipelines = list(filter(lambda x: x.id == pipeline_id, pipelines))
        if len(found_pipelines) == 0:
            return None

        if len(found_pipelines) > 1:
            raise Exception(
                "Found more than 1 pipeline with this ID.  This shouldn't happen."
            )

        return found_pipelines[0]




from __future__ import annotations
from typing import List, Dict, Optional, Any
import os
import json
from tqdm import tqdm
from tqdm.utils import CallbackIOWrapper
import requests.exceptions
import requests
from tonic_textual.classes.tonic_exception import DatasetFileMatchesExistingFile
from tonic_textual.classes.httpclient import HttpClient
from tonic_textual.classes.datasetfile import DatasetFile
from tonic_textual.services.datasetfile import DatasetFileService


class Dataset:
    """Class to represent and provide access to a Tonic Textual dataset.

    Parameters
    ----------
    id: str
        Dataset identifier.

    name: str
        Dataset name.

    files: Dict
        Serialized DatasetFile objects representing the files in a dataset.

    client: HttpClient
        The HTTP client to use.
    """

    def __init__(
        self, id: str, name: str, files: List[Dict[str, Any]], client: HttpClient
    ):
        self.id = id
        self.name = name
        self.client = client
        self.datasetfile_service = DatasetFileService(self.client)
        self.files = [
            DatasetFile(
                f["fileId"],
                f["fileName"],
                f.get("numRows"),
                f["numColumns"],
                f["processingStatus"],
                f.get("processingError"),
            )
            for f in files
        ]

        if len(self.files) > 0:
            self.num_columns = max([f.num_columns for f in self.files])
        else:
            self.num_columns = None

        self.num_rows_per_request = 1000
        self.total_rows_fetched = 0

        self._total_rows_fetched_in_current_file = 0
        self._cur_file_idx = 0

    def reset(self):
        self.total_rows_fetched = 0
        self._cur_file_idx = 0
        self._total_rows_fetched_in_current_file = 0

    def upload_then_add_file(self, file_path: str, file_name: Optional[str] = None):
        """
        Uploads a file to the dataset.

        Parameters
        --------
        file_path: str
            The absolute path of the file to upload.
        file_name: str
            The name of the file to save to Tonic Textual.

        Raises
        ------

        DatasetFileMatchesExistingFile
            Returned if the file content matches an existing file.

        """
        if file_name is None:
            file_name = os.path.basename(file_path)
        file_size = os.path.getsize(file_path)
        with open(file_path, "rb") as f:
            with tqdm(
                desc="[INFO] Uploading",
                total=file_size,
                unit="B",
                unit_scale=True,
                unit_divisor=1024,
            ) as t:
                reader_wrapper = CallbackIOWrapper(t.update, f, "read")

                files = {
                    "document": (
                        None,
                        json.dumps(
                            {
                                "fileName": file_name,
                                "csvConfig": {},
                                "datasetId": self.id,
                            }
                        ),
                        "application/json",
                    ),
                    "file": reader_wrapper,
                }
                try:
                    updated_dataset = self.client.http_post(
                        f"/api/dataset/{self.id}/files/upload", files=files
                    )
                    # numRows is null when a file is first uploaded
                except requests.exceptions.HTTPError as e:
                    if e.response.status_code == 409:
                        raise DatasetFileMatchesExistingFile(e)
                    else:
                        raise e
        self.files = [
            DatasetFile(
                f["fileId"],
                f["fileName"],
                f.get("numRows"),
                f["numColumns"],
                f["processingStatus"],
                f.get("processingError"),
            )
            for f in updated_dataset["files"]
        ]
        self.num_columns = max([f.num_columns for f in self.files])

    def fetch_all_df(self):
        """
        Fetches all of the data in the dataset as a pandas dataframe.

        Returns
        -------
        pd.DataFrame
            Dataset data in a pandas dataframe.
        """
        try:
            import pandas as pd
        except ImportError as e:
            raise ImportError(
                "Pandas is required to fetch the dataset data as a pandas dataframe. Please install pandas before using this method."
            ) from e
        data = self._fetch_all()

        if self.num_columns is None:
            return pd.DataFrame()

        # RAW file, not CSV
        if self.num_columns == 0:
            if len(data) == 0:
                return pd.DataFrame(columns=["text"])
            return pd.DataFrame(data, columns=["text"])

        columns = ["col" + str(x) for x in range(self.num_columns)]
        if len(data) == 0:
            return pd.DataFrame(columns=columns)
        else:
            return pd.DataFrame(data, columns=columns)

    def fetch_all_json(self) -> str:
        """
        Fetches all of the data in the dataset as JSON.

        Returns
        -------
        str
            Dataset data in JSON format.
        """
        return json.dumps(self._fetch_all())

    def _fetch_all(self) -> List[List[str]]:
        """
        Fetches all data from the dataset.

        Returns
        -------
        List[List[str]]
            The datset data.
        """
        response = []
        with requests.Session() as session:
            for file in self.files:
                try:
                    if file.num_columns == 0:
                        more_data = self.client.http_get_file(
                            f"/api/dataset/{self.id}/files/{file.id}/get_data",
                            session=session,
                        )
                        response += [[more_data]]
                    else:
                        more_data = self.client.http_get(
                            f"/api/dataset/{self.id}/files/{file.id}/get_data",
                            session=session,
                        )
                        response += more_data
                except requests.exceptions.HTTPError as e:
                    if e.response.status_code == 409:
                        continue
                    else:
                        raise e
            return response

    def get_processed_files(self) -> List[DatasetFile]:
        """
        Gets all of the files in the dataset for which processing is complete. The data
        in these files is returned when data is requested.

        Returns
        ------
        List[DatasetFile]:
            The list of processed dataset files.
        """
        return list(filter(lambda x: x.processing_status == "Completed", self.files))

    def get_queued_files(self) -> List[DatasetFile]:
        """
        Gets all of the files in the dataset that are waiting to be processed.

        Returns
        ------
        List[DatasetFile]:
            The list of dataset files that await processing.
        """
        return list(filter(lambda x: x.processing_status == "Queued", self.files))

    def get_running_files(self) -> List[DatasetFile]:
        """
        Gets all of the files in the dataset that are currently being processed.

        Returns
        ------
        List[DatasetFile]:
            The list of files that are being processed.
        """
        return list(filter(lambda x: x.processing_status == "Running", self.files))

    def get_failed_files(self) -> List[DatasetFile]:
        """
        Gets all of the files in dataset that encountered an error when they were
        processed. These files are effectively ignored.

        Returns
        ------
        List[DatasetFile]:
            The list of files that had processing errors.
        """
        return list(filter(lambda x: x.processing_status == "Failed", self.files))

    def _check_processing_and_update(self):
        """
        Checks the processing status of the files in the dataset and updates the files
        list.
        """
        if len(self.get_queued_files() + self.get_running_files()) > 0:
            self.files = self.datasetfile_service.get_files(self.id)

    def describe(self) -> str:
        """
        Returns a string of the dataset name, identifier, and the list of files.

        Examples
        --------
        >>> workspace.describe()
        Dataset: your_dataset_name [dataset_id]
        Number of Files: 2
        Number of Rows: 1000
        """
        self._check_processing_and_update()

        files_waiting_for_proc = self.get_queued_files() + self.get_running_files()
        files_with_error = self.get_failed_files()
        number_rows = sum(
            [x.num_rows if x.num_rows is not None else 0 for x in self.files]
        )
        result = f"Dataset: {self.name} [{self.id}]\n"
        result += f"Number of Files: {len(self.get_processed_files())}\n"
        result += "Files that are waiting for processing: "
        result += f"{', '.join([str((f.id,f.name)) for f in files_waiting_for_proc])}\n"
        result += "Files that encountered errors while processing: "
        result += f"{', '.join([str((f.id,f.name)) for f in files_with_error])}\n"
        result += f"Number of Rows: {number_rows}\n"
        result += f"Number of rows fetched: {self.total_rows_fetched}"
        return result



from typing import List, Optional

import requests
import json
import io

from requests import RequestException

from tonic_textual.classes.SolarCsvConfig import SolarCsvConfig
from tonic_textual.classes.file_parse_result_diff_enumerator import (
    FileParseResultsDiffEnumerator,
)
from tonic_textual.classes.httpclient import HttpClient
from tonic_textual.classes.pipeline_file_enumerator import PipelineFileEnumerator
from tonic_textual.classes.pipeline_run import PipelineRun
from tonic_textual.classes.tonic_exception import FileUploadError


class Pipeline:
    """Class to represent and provide access to a Tonic Textual pipeline.

    Parameters
    ----------
    name: str
        Pipeline name.

    id: str
        Pipeline identifier.

    client: HttpClient
        The HTTP client to use.
    """

    def __init__(self, name: str, id: str, client: HttpClient):
        self.id = id
        self.name = name
        self.client = client

    def describe(self) -> str:
        """Returns the name and id of the pipeline."""
        description = "--------------------------------------------------------\n"
        description += f"Name: {self.name}\n"
        description += f"ID: {self.id}\n"
        description += "--------------------------------------------------------\n"
        return description

    def get_runs(self) -> List[PipelineRun]:
        """Get the runs for the pipeline.

        Returns
        -------
        List[PipelineRun]
            A list of PipelineRun objects.
        """
        with requests.Session() as session:
            response = self.client.http_get(
                f"/api/parsejob/{self.id}/jobs", session=session
            )
            runs: List[PipelineRun] = []
            for run in response:
                runs.append(
                    PipelineRun(run["id"], run["status"], run["endTime"], self.client)
                )
            return runs

    def run(self) -> str:
        """Run the pipeline.

        Returns
        -------
        str
            The ID of the job.
        """
        try:
            response = self.client.http_post(
                f"/api/parsejobconfig/{self.id}/start"
            )
            return response
        except RequestException as req_err:
            if hasattr(req_err, 'response') and req_err.response is not None:
                status_code = req_err.response.status_code
                error_message = req_err.response.text
                raise FileUploadError(f"Error {status_code}: {error_message}")
            else:
                raise req_err


    def enumerate_files(self, lazy_load_content=True) -> PipelineFileEnumerator:
        """Enumerate the files in the pipeline.

        Parameters
        ----------
        lazy_load_content: bool
            Whether to lazily load the content of the files. Default is True.

        Returns
        -------
        PipelineFileEnumerator
            An enumerator for the files in the pipeline.
        """
        runs = self.get_runs()
        successful_runs = filter(lambda r: r.status == "Completed", runs)
        sorted_finished_runs = sorted(
            successful_runs, key=lambda r: r.end_time, reverse=True
        )

        if len(sorted_finished_runs) == 0:
            return PipelineFileEnumerator(
                "", self.client, lazy_load_content=lazy_load_content
            )

        job_id = sorted_finished_runs[0].id
        return PipelineFileEnumerator(
            job_id, self.client, lazy_load_content=lazy_load_content
        )

    def get_delta(
        self, pipeline_run1: PipelineRun, pipeline_run2: PipelineRun
    ) -> FileParseResultsDiffEnumerator:
        """Enumerates the files in the diff between two pipeline runs.

        Parameters
        ----------
        pipeline_run1: PipelineRun
            The first pipeline run.

        pipeline_run2: PipelineRun
            The second pipeline run.

        Returns
        -------
        FileParseResultsDiffEnumerator
            An enumerator for the files in the diff between the two runs.
        """
        return FileParseResultsDiffEnumerator(
            pipeline_run1.id, pipeline_run2.id, self.client
        )

    def upload_file(self, file: io.IOBase, file_name: str,
                    csv_config: Optional[SolarCsvConfig] = None) -> str:
        """Upload a file to the pipeline.

        Parameters
        ----------
        pipeline_id: str
            The ID of the pipeline.
        file: io.IOBase
            The file to upload.
        file_name: str
            The name of the file.
        csv_config: SolarCsvConfig
            The configuration for the CSV file. This is optional.

        Returns
        -------
        None
            This function does not return any value.
        """
        files = {
            "document": (
                None,
                json.dumps({
                    "fileName": file_name,
                    "csvConfig": csv_config
                }),
                "application/json",
            ),
            "file": (file_name, file, "application/octet-stream"),
        }

        try:
            return self.client.http_post(f"/api/parsejobconfig/{self.id}/local-files/upload", files=files)
        except RequestException as req_err:
            if hasattr(req_err, 'response') and req_err.response is not None:
                status_code = req_err.response.status_code
                error_message = req_err.response.text
                raise FileUploadError(f"Error {status_code}: {error_message}")
            else:
                raise req_err

